﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Security.Principal;

#if !NETSTANDARD
using System.Security.Permissions;
#endif

namespace ImpersonateCLI
{
    /// <summary>
    /// Provides ability to run code within the context of a specific user.
    /// </summary>
#if !NETSTANDARD
    [PermissionSet(SecurityAction.Demand, Name = "FullTrust")]
#endif
    public static class Impersonation
    {



        /// <summary>
        /// Impersonates a specific user account to execute the specified function.
        /// </summary>
        /// <typeparam name="T">The return type of the function.</typeparam>
        /// <param name="credentials">The credentials of the user account to impersonate.</param>
        /// <param name="logonType">The logon type used when impersonating the user account.</param>
        /// <param name="function">The function to execute, which accepts a <see cref="SafeAccessTokenHandle"/> to the user account as its only parameter.</param>
        /// <returns>The result of executing the function.</returns>
        public static T RunAsUser<T>(UserCredentials credentials, LogonType logonType, Func<T> function)
        {
            using (var tokenHandle = credentials.Impersonate(logonType))
            {
                return RunImpersonated(tokenHandle, _ => function());
            }
        }

        /// <summary>
        /// Impersonates a specific user account to execute the specified function.
        /// </summary>
        /// <typeparam name="T">The return type of the function.</typeparam>
        /// <param name="credentials">The credentials of the user account to impersonate.</param>
        /// <param name="logonType">The logon type used when impersonating the user account.</param>
        /// <param name="function">The function to execute.</param>
        /// <returns>The result of executing the function.</returns>
#pragma warning disable CS0436 // Type conflicts with imported type
        public static T RunAsUser<T>(UserCredentials credentials, LogonType logonType, Func<SafeAccessTokenHandle, T> function)
#pragma warning restore CS0436 // Type conflicts with imported type
        {
            using (var tokenHandle = credentials.Impersonate(logonType))
            {
                return RunImpersonated(tokenHandle, function);
            }
        }

#pragma warning disable CS0436 // Type conflicts with imported type
        private static T RunImpersonated<T>(SafeAccessTokenHandle tokenHandle, Func<SafeAccessTokenHandle, T> function)
#pragma warning restore CS0436 // Type conflicts with imported type
        {
#if NETSTANDARD || NET46
            return WindowsIdentity.RunImpersonated(tokenHandle, () => function(tokenHandle));
#else
            // ReSharper disable once UnusedVariable
            using (var context = WindowsIdentity.Impersonate(tokenHandle.DangerousGetHandle()))
            {
                return function(tokenHandle);
            }
#endif
        }
    }
}